/*
    Simplified user-agent stylesheet for HTML5 in tests.
*/
html, body, div, h1, h2, h3, h4, ol, p, ul, hr, pre, section, article
                { display: block; }
li              { display: list-item }
head            { display: none }
pre             { white-space: pre }
br:before       { content: '\A'; white-space: pre-line }
ol              { list-style-type: decimal }
ol, ul          { counter-reset: list-item }

table           { display: table;
                  box-sizing: border-box }
tr              { display: table-row }
thead           { display: table-header-group }
tbody           { display: table-row-group }
tfoot           { display: table-footer-group }
col             { display: table-column }
colgroup        { display: table-column-group }
td, th          { display: table-cell }
caption         { display: table-caption }

a[href] { -weasy-link: attr(href); }
a[name] { -weasy-anchor: attr(name); }
*[id] { -weasy-anchor: attr(id); }
h1 { -weasy-bookmark-level: 1; -weasy-bookmark-label: contents; }
h2 { -weasy-bookmark-level: 2; -weasy-bookmark-label: contents; }
h3 { -weasy-bookmark-level: 3; -weasy-bookmark-label: contents; }
h4 { -weasy-bookmark-level: 4; -weasy-bookmark-label: contents; }
h5 { -weasy-bookmark-level: 5; -weasy-bookmark-label: contents; }
h6 { -weasy-bookmark-level: 6; -weasy-bookmark-label: contents; }
