# coding: utf8
"""
    weasyprint.tests.w3_test_suite
    ------------------------------

    Everything related to the W3C CSS 2.1 Test Suite.

    See http://test.csswg.org/suites/css2.1/

    :copyright: Copyright 2011-2012 Simon Sapin and contributors, see AUTHORS.
    :license: BSD, see LICENSE for details.

"""

from __future__ import division, unicode_literals
